class GraphStructuredStack(object):
    """
    A data structure representing a forkable stack from Tomita's paper.
    """


class PackedParseForestNode(object):
    """This is a node in the Tomita's packed parse forest."""


class NonTerminal(PackedParseForestNode):
    pass


class Terminal(PackedParseForestNode):
    pass
